package pathfinder;

import java.util.ArrayList;
import java.util.Collections;

public class AStar extends PathFinder {
    
    private int maxListSize = 1<<6;
    
    // Profiler shows significant overhead due to arraylists
    // Bad things will happens when more than 1000 elements go into these lists
    // However that probably won't happen
    Node[]
            openList = new Node[maxListSize],
            closedList = new Node[maxListSize];
    private int openCount = 0;
    private int closedCount = 0;
    
    private final static int DIR_COUNT = 8;
    private final static float sqrt2 = (float)Math.sqrt(2);
    private final static int[]
            // Directions:
            // r=right, u=up, d=down, l=left
            //        r  u  l  d ur ul dl dr
            DIR_X = { 1, 0,-1, 0, 1,-1,-1, 1},
            DIR_Y = { 0, 1, 0,-1, 1, 1,-1,-1};
    private final static float[] COST  = {1f,1f,1f,1f, sqrt2, sqrt2, sqrt2, sqrt2};

    private int iterationCount = 0;
    
    final static int UNKNOWN = -1;
    
    Node best = null;
    
    Node src, dst;
    private PathClient client;
    
    final Node[] nodes;

    private boolean running = false;
    
    public AStar(int width, int height){
        super(width, height);
        this.nodes = new Node[width*height];
        for(int i = 0; i < width; i++){
            for(int j = 0; j < height; j++){
                nodes[i+width*j] = new Node(i,j);
            }
        }
    }

    public void clear(){
        for(int i = 0; i < openCount; i++){
            openList[i].reset();
        }
        for(int i = 0; i < closedCount; i++){
            closedList[i].reset();
        }
        openCount = 0;
        closedCount = 0;
        best = null;
    }
    
    public boolean setup(int srcX, int srcY, int dstX, int dstY, PathClient client){
        if(client.isBlocked(dstX, dstY)){
            return false;
        }
        clear();
        this.client = client;
        
        src = nodes[srcX + width*srcY];
        dst = nodes[dstX + width*dstY];
        int distance = dist(src, dst);
        src.setValues(distance, 0);
        openList[openCount++] = src;
        src.inOpenList = true;
        iterationCount = 0;
        running = true;
        return true;
    }
    
    public int findPath(int iterations){
        int status = RUNNING;
        for(int i = 0; i < iterations && status == RUNNING; i++){
            status = iterate();
        }
        return status;
    }

    public int iterate(){
        if(!running || openCount==0){
            return ERROR;
        }
        
        if(openList.length < 2*DIR_COUNT + openCount){
            Node[] old = openList;
            openList = new Node[2*openList.length];
            System.arraycopy(old, 0, openList, 0, openCount);
        }
        
        if(closedList.length < 2*DIR_COUNT + closedCount){
            Node[] old = closedList;
            closedList = new Node[2*closedList.length];
            System.arraycopy(old, 0, closedList, 0, closedCount);
        }
        
        best = openList[--openCount];
        
        best.inOpenList = false;
        best.inClosedList = true;
        
        closedList[closedCount++] = best;

        if(best == dst){
            running = false;
            return COMPLETED;
        }
        iterationCount++;
        for(int i = 0; i < DIR_COUNT; i++){
            int x = best.x + DIR_X[i];
            int y = best.y + DIR_Y[i];
            
            boolean blocked = client.isBlocked(x, y);
            if(!blocked){
                check(best, nodes[x + width*y], COST[i]*client.getCost(x,y));                
            }
        }
        return RUNNING;
    }
    
    public void check(Node parent, Node other, float cost){
        float g = parent.g + cost;
        if(other.inOpenList || other.inClosedList){
            if(g < other.g){
                other.parent = parent;
                other.g = g;
                other.f = g + other.h;
            }
        } else {
            other.parent = parent;

            int dist = dist(other, dst);
            
            other.g = g;
            other.h = dist;
            other.f = g + dist;

            int index = java.util.Arrays.binarySearch(openList, 0, openCount, other);
            
            if(index < 0){
                index = - (index + 1);
            }
            other.inOpenList = true;
            openCount++;

            System.arraycopy(openList, index, openList, index+1, openCount-index);
            
            openList[index] = other;
        }
    }
    
    public ArrayList<Node> getPath(){
        Node node = best;
        ArrayList<Node> path = new ArrayList<Node>();
        while(node != null){
            path.add(node);
            node = node.parent;
        }
        //running = false;
        Collections.reverse(path);
        return path;
    }
    
    private int dist(Node a, Node b){
        float dx = b.x - a.x;
        float dy = b.y - a.y;
        //Math.hypot makes this thing take >3 times as long for some reason!!
        //This is as fast as return dx*dx + dy*dy in practice
        return (int)(8f*Math.sqrt(dx*dx + dy*dy));
        //return (int)(8f*Math.hypot(dx, dy));
    }

    public int getIterationCount(){
        return iterationCount;
    }
    
    public int getNodeCount(){
        return width*height;
    }
    
    public Node getNode(int index){
        return nodes[index];
    }
    
    public Node getNode(int x, int y){
        return nodes[x + y*width];
    }
    
    public Node getSrc(){
        return src;
    }
    
    public Node getDst(){
        return dst;
    }
}
